#ifndef G2O_PY_H
#define G2O_PY_H

#include <pybind11/eigen.h>
#include <pybind11/operators.h>
#include <pybind11/pybind11.h>
#include <pybind11/stl.h>

#include <memory>

namespace py = pybind11;             // NOLINT
using namespace pybind11::literals;  // NOLINT

#endif
